"""mateqn_test.py - test suite for matrix equation solvers

Copyright (c) 2020, All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the name of the project author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Author: Bjorn Olofsson
"""

import numpy as np
from numpy import array, zeros
from numpy.testing import assert_array_almost_equal, assert_array_less
import pytest
from scipy.linalg import eigvals, solve

import control as ct
from control.mateqn import lyap, dlyap, care, dare
from control.exception import ControlArgument, ControlDimension, slycot_check
from control.tests.conftest import slycotonly


class TestMatrixEquations:
    """These are tests for the matrix equation solvers in mateqn.py"""

    def test_lyap(self):
        A = array([[-1, 1], [-1, 0]])
        Q = array([[1, 0], [0, 1]])
        X = lyap(A, Q)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X + X @ A.T + Q, zeros((2,2)))

        A = array([[1, 2], [-3, -4]])
        Q = array([[3, 1], [1, 1]])
        X = lyap(A,Q)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X + X @ A.T + Q, zeros((2,2)))

        # Compare methods
        if slycot_check():
            X_scipy = lyap(A, Q, method='scipy')
            X_slycot = lyap(A, Q, method='slycot')
            assert_array_almost_equal(X_scipy, X_slycot)

    def test_lyap_sylvester(self):
        A = 5
        B = array([[4, 3], [4, 3]])
        C = array([2, 1])
        X = lyap(A, B, C)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A * X + X @ B + C, zeros((1,2)))

        A = array([[2, 1], [1, 2]])
        B = array([[1, 2], [0.5, 0.1]])
        C = array([[1, 0], [0, 1]])
        X = lyap(A, B, C)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X + X @ B + C, zeros((2,2)))

        # Compare methods
        if slycot_check():
            X_scipy = lyap(A, B, C, method='scipy')
            X_slycot = lyap(A, B, C, method='slycot')
            assert_array_almost_equal(X_scipy, X_slycot)

    @slycotonly
    def test_lyap_g(self):
        A = array([[-1, 2], [-3, -4]])
        Q = array([[3, 1], [1, 1]])
        E = array([[1, 2], [2, 1]])
        X = lyap(A, Q, None, E)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X @ E.T + E @ X @ A.T + Q,
                                  zeros((2,2)))

        # Make sure that trying to solve with SciPy generates an error
        with pytest.raises(ControlArgument, match="'scipy' not valid"):
            X = lyap(A, Q, None, E, method='scipy')

    def test_dlyap(self):
        A = array([[-0.6, 0],[-0.1, -0.4]])
        Q = array([[1,0],[0,1]])
        X = dlyap(A,Q)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X @ A.T - X + Q, zeros((2,2)))

        A = array([[-0.6, 0],[-0.1, -0.4]])
        Q = array([[3, 1],[1, 1]])
        X = dlyap(A,Q)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X @ A.T - X + Q, zeros((2,2)))

    @slycotonly
    def test_dlyap_g(self):
        A = array([[-0.6, 0],[-0.1, -0.4]])
        Q = array([[3, 1],[1, 1]])
        E = array([[1, 1],[2, 1]])
        X = dlyap(A, Q, None, E)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X @ A.T - E @ X @ E.T + Q,
                                  zeros((2,2)))

        # Make sure that trying to solve with SciPy generates an error
        with pytest.raises(ControlArgument, match="'scipy' not valid"):
            X = dlyap(A, Q, None, E, method='scipy')

    @slycotonly
    def test_dlyap_sylvester(self):
        A = 5
        B = array([[4, 3], [4, 3]])
        C = array([2, 1])
        X = dlyap(A,B,C)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A * X @ B.T - X + C, zeros((1,2)))

        A = array([[2, 1], [1, 2]])
        B = array([[1, 2], [0.5, 0.1]])
        C = array([[1, 0], [0, 1]])
        X = dlyap(A, B, C)
        # print("The solution obtained is ", X)
        assert_array_almost_equal(A @ X @ B.T - X + C, zeros((2,2)))

        # Make sure that trying to solve with SciPy generates an error
        with pytest.raises(ControlArgument, match="'scipy' not valid"):
            X = dlyap(A, B, C, method='scipy')

    def test_care(self):
        A = array([[-2, -1],[-1, -1]])
        Q = array([[0, 0],[0, 1]])
        B = array([[1, 0],[0, 4]])

        X, L, G = care(A, B, Q)
        # print("The solution obtained is", X)
        M = A.T @ X + X @ A - X @ B @ B.T @ X + Q
        assert_array_almost_equal(M,
                                  zeros((2,2)))
        assert_array_almost_equal(B.T @ X, G)

        # Compare methods
        if slycot_check():
            X_scipy, L_scipy, G_scipy = care(A, B, Q, method='scipy')
            X_slycot, L_slycot, G_slycot = care(A, B, Q, method='slycot')
            assert_array_almost_equal(X_scipy, X_slycot)
            assert_array_almost_equal(np.sort(L_scipy), np.sort(L_slycot))
            assert_array_almost_equal(G_scipy, G_slycot)

    def test_care_g(self):
        A = array([[-2, -1],[-1, -1]])
        Q = array([[0, 0],[0, 1]])
        B = array([[1, 0],[0, 4]])
        R = array([[2, 0],[0, 1]])
        S = array([[0, 0],[0, 0]])
        E = array([[2, 1],[1, 2]])

        X,L,G = care(A,B,Q,R,S,E)
        # print("The solution obtained is", X)
        Gref = solve(R, B.T @ X @ E + S.T)
        assert_array_almost_equal(Gref, G)
        assert_array_almost_equal(
            A.T @ X @ E + E.T @ X @ A
            - (E.T @ X @ B + S) @ Gref + Q,
            zeros((2,2)))

        # Compare methods
        if slycot_check():
            X_scipy, L_scipy, G_scipy = care(
                A, B, Q, R, S, E, method='scipy')
            X_slycot, L_slycot, G_slycot = care(
                A, B, Q, R, S, E, method='slycot')
            assert_array_almost_equal(X_scipy, X_slycot)
            assert_array_almost_equal(np.sort(L_scipy), np.sort(L_slycot))
            assert_array_almost_equal(G_scipy, G_slycot)

    def test_care_g2(self):
        A = array([[-2, -1],[-1, -1]])
        Q = array([[0, 0],[0, 1]])
        B = array([[1],[0]])
        R = 1
        S = array([[1],[0]])
        E = array([[2, 1],[1, 2]])

        X,L,G = care(A,B,Q,R,S,E)
        # print("The solution obtained is", X)
        Gref = 1/R * (B.T @ X @ E + S.T)
        assert_array_almost_equal(
            A.T @ X @ E + E.T @ X @ A
            - (E.T @ X @ B + S) @ Gref + Q ,
            zeros((2,2)))
        assert_array_almost_equal(Gref , G)

        # Compare methods
        if slycot_check():
            X_scipy, L_scipy, G_scipy = care(
                A, B, Q, R, S, E, method='scipy')
            X_slycot, L_slycot, G_slycot = care(
                A, B, Q, R, S, E, method='slycot')
            assert_array_almost_equal(X_scipy, X_slycot)
            assert_array_almost_equal(L_scipy, L_slycot)
            assert_array_almost_equal(G_scipy, G_slycot)

    def test_dare(self):
        A = array([[-0.6, 0],[-0.1, -0.4]])
        Q = array([[2, 1],[1, 0]])
        B = array([[2, 1],[0, 1]])
        R = array([[1, 0],[0, 1]])

        X, L, G = dare(A, B, Q, R)
        # print("The solution obtained is", X)
        Gref = solve(B.T @ X @ B + R, B.T @ X @ A)
        assert_array_almost_equal(Gref, G)
        assert_array_almost_equal(
            X, A.T @ X @ A - A.T @ X @ B @ Gref + Q)
        # check for stable closed loop
        lam = eigvals(A - B @ G)
        assert_array_less(abs(lam), 1.0)

        A = array([[1, 0],[-1, 1]])
        Q = array([[0, 1],[1, 1]])
        B = array([[1],[0]])
        R = 2

        X, L, G = dare(A, B, Q, R)
        # print("The solution obtained is", X)
        AtXA = A.T @ X @ A
        AtXB = A.T @ X @ B
        BtXA = B.T @ X @ A
        BtXB = B.T @ X @ B
        assert_array_almost_equal(
            X, AtXA - AtXB @ solve(BtXB + R, BtXA) + Q)
        assert_array_almost_equal(BtXA / (BtXB + R), G)
        # check for stable closed loop
        lam = eigvals(A - B @ G)
        assert_array_less(abs(lam), 1.0)

    def test_dare_compare(self):
        A = np.array([[-0.6, 0], [-0.1, -0.4]])
        Q = np.array([[2, 1], [1, 0]])
        B = np.array([[2, 1], [0, 1]])
        R = np.array([[1, 0], [0, 1]])
        S = np.zeros((A.shape[0], B.shape[1]))
        E = np.eye(A.shape[0])

        # Solve via scipy
        X_scipy, L_scipy, G_scipy = dare(A, B, Q, R, method='scipy')

        # Solve via slycot
        if ct.slycot_check():
            X_slicot, L_slicot, G_slicot = dare(
                A, B, Q, R, S, E, method='scipy')
            np.testing.assert_almost_equal(X_scipy, X_slicot)
            np.testing.assert_almost_equal(L_scipy, L_slicot)
            np.testing.assert_almost_equal(G_scipy, G_slicot)

    def test_dare_g(self):
        A = array([[-0.6, 0],[-0.1, -0.4]])
        Q = array([[2, 1],[1, 3]])
        B = array([[1, 5],[2, 4]])
        R = array([[1, 0],[0, 1]])
        S = array([[1, 0],[2, 0]])
        E = array([[2, 1],[1, 2]])

        X, L, G = dare(A, B, Q, R, S, E)
        # print("The solution obtained is", X)
        Gref = solve(B.T @ X @ B + R, B.T @ X @ A + S.T)
        assert_array_almost_equal(Gref, G)
        assert_array_almost_equal(
            E.T @ X @ E,
            A.T @ X @ A - (A.T @ X @ B + S) @ Gref + Q)
        # check for stable closed loop
        lam = eigvals(A - B @ G, E)
        assert_array_less(abs(lam), 1.0)

    def test_dare_g2(self):
        A = array([[-0.6, 0], [-0.1, -0.4]])
        Q = array([[2, 1], [1, 3]])
        B = array([[1], [2]])
        R = 1
        S = array([[1], [2]])
        E = array([[2, 1], [1, 2]])

        X, L, G = dare(A, B, Q, R, S, E)
        # print("The solution obtained is", X)
        AtXA = A.T @ X @ A
        AtXB = A.T @ X @ B
        BtXA = B.T @ X @ A
        BtXB = B.T @ X @ B
        EtXE = E.T @ X @ E
        assert_array_almost_equal(
            EtXE, AtXA - (AtXB + S)  @ solve(BtXB + R, BtXA + S.T) + Q)
        assert_array_almost_equal((BtXA + S.T) / (BtXB + R), G)
        # check for stable closed loop
        lam = eigvals(A - B @ G, E)
        assert_array_less(abs(lam), 1.0)

    def test_raise(self):
        """ Test exception raise for invalid inputs """

        # correct shapes and forms
        A = array([[1, 0], [-1, -1]])
        Q = array([[2, 1], [1, 2]])
        C = array([[1, 0], [0, 1]])
        E = array([[2, 1], [1, 2]])

        # these fail
        Afq = array([[1, 0, 0], [-1, -1, 0]])
        Qfq = array([[2, 1, 0], [1, 2, 0]])
        Qfs = array([[2, 1], [-1, 2]])
        Cfd = array([[1, 0, 0], [0, 1, 0]])
        Efq = array([[2, 1, 0], [1, 2, 0]])

        for cdlyap in [lyap, dlyap]:
            with pytest.raises(ControlDimension):
                cdlyap(Afq, Q)
            with pytest.raises(ControlDimension):
                cdlyap(A, Qfq)
            with pytest.raises(ControlArgument):
                cdlyap(A, Qfs)
            with pytest.raises(ControlDimension):
                cdlyap(Afq, Q, C)
            with pytest.raises(ControlDimension):
                cdlyap(A, Qfq, C)
            with pytest.raises(ControlDimension):
                cdlyap(A, Q, Cfd)
            with pytest.raises(ControlDimension):
                cdlyap(A, Qfq, None, E)
            with pytest.raises(ControlDimension):
                cdlyap(A, Q, None, Efq)
            with pytest.raises(ControlArgument):
                cdlyap(A, Qfs, None, E)
            with pytest.raises(ControlArgument):
                cdlyap(A, Q, C, E)

        B = array([[1, 0], [0, 1]])
        Bf = array([[1, 0], [0, 1], [1, 1]])
        R = Q
        Rfs = Qfs
        Rfq = Qfq
        S = array([[0, 0], [0, 0]])
        Sf = array([[0, 0, 0], [0, 0, 0]])
        E = array([[2, 1], [1, 2]])
        Ef = array([[2, 1], [1, 2], [1, 2]])

        with pytest.raises(ControlDimension):
            care(Afq, B, Q)
        with pytest.raises(ControlDimension):
            care(A, B, Qfq)
        with pytest.raises(ControlDimension):
            care(A, Bf, Q)
        with pytest.raises(ControlDimension):
            care(1, B, 1)
        with pytest.raises(ControlArgument):
            care(A, B, Qfs)
        with pytest.raises(ControlArgument):
            dare(A, B, Q, Rfs)
        for cdare in [care, dare]:
            with pytest.raises(ControlDimension):
                cdare(Afq, B, Q, R, S, E)
            with pytest.raises(ControlDimension):
                cdare(A, B, Qfq, R, S, E)
            with pytest.raises(ControlDimension):
                cdare(A, Bf, Q, R, S, E)
            with pytest.raises(ControlDimension):
                cdare(A, B, Q, R, S, Ef)
            with pytest.raises(ControlDimension):
                cdare(A, B, Q, Rfq, S, E)
            with pytest.raises(ControlDimension):
                cdare(A, B, Q, R, Sf, E)
            with pytest.raises(ControlArgument):
                cdare(A, B, Qfs, R, S, E)
            with pytest.raises(ControlArgument):
                cdare(A, B, Q, Rfs, S, E)



"""input_element_int_test.py

Author: Kangwon Lee (kangwonlee)
Date: 22 Oct 2017

Modified:
* 29 Dec 2017, RMM - updated file name and added header
"""

import numpy as np
from control import dcgain, ss, tf

class TestTfInputIntElement:
    """input_element_int_test

    Unit tests contributed as part of PR gh-158, "SISO tf() may not work
    with numpy arrays with numpy.int elements
    """

    def test_tf_den_with_numpy_int_element(self):
        num = 1
        den = np.convolve([1, 2, 1], [1, 1, 1])

        sys = tf(num, den)

        np.testing.assert_almost_equal(1., dcgain(sys))

    def test_tf_num_with_numpy_int_element(self):
        num = np.convolve([1], [1, 1])
        den = np.convolve([1, 2, 1], [1, 1, 1])

        sys = tf(num, den)

        np.testing.assert_almost_equal(1., dcgain(sys))

    # currently these pass
    def test_tf_input_with_int_element(self):
        num = 1
        den = np.convolve([1.0, 2, 1], [1, 1, 1])

        sys = tf(num, den)

        np.testing.assert_almost_equal(1., dcgain(sys))

    def test_ss_input_with_int_element(self):
        a = np.array([[0, 1],
                      [-1, -2]], dtype=int)
        b = np.array([[0],
                       [1]], dtype=int)
        c = np.array([[0, 1]], dtype=int)
        d = np.array([[1]], dtype=int)

        sys = ss(a, b, c, d)
        sys2 = tf(sys)
        np.testing.assert_almost_equal(dcgain(sys), dcgain(sys2))

    def test_ss_input_with_0int_dcgain(self):
        a = np.array([[0, 1],
                      [-1, -2]], dtype=int)
        b = np.array([[0],
                       [1]], dtype=int)
        c = np.array([[0, 1]], dtype=int)
        d = 0
        sys = ss(a, b, c, d)
        np.testing.assert_allclose(dcgain(sys), 0,
                                   atol=np.finfo(float).epsneg)



# rlocus.py - code for computing a root locus plot
# Code contributed by Ryan Krauss, 2010
#
# Copyright (c) 2010 by Ryan Krauss
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# RMM, 17 June 2010: modified to be a standalone piece of code
#   * Added BSD copyright info to file (per Ryan)
#   * Added code to convert (num, den) to poly1d's if they aren't already.
#     This allows Ryan's code to run on a standard signal.ltisys object
#     or a control.TransferFunction object.
#   * Added some comments to make sure I understand the code
#
# RMM, 2 April 2011: modified to work with new LTI structure (see ChangeLog)
#   * Not tested: should still work on signal.ltisys objects
#
# Sawyer B. Fuller (minster@uw.edu) 21 May 2020:
#   * added compatibility with discrete-time systems.
#
# $Id$

# Packages used by this module
from functools import partial
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from numpy import array, poly1d, row_stack, zeros_like, real, imag
import scipy.signal             # signal processing toolbox
from .namedio import isdtime
from .xferfcn import _convert_to_transfer_function
from .exception import ControlMIMONotImplemented
from .sisotool import _SisotoolUpdate
from .grid import sgrid, zgrid
from . import config
import warnings

__all__ = ['root_locus', 'rlocus']

# Default values for module parameters
_rlocus_defaults = {
    'rlocus.grid': True,
    'rlocus.plotstr': 'b' if int(mpl.__version__[0]) == 1 else 'C0',
    'rlocus.print_gain': True,
    'rlocus.plot': True
}


# Main function: compute a root locus diagram
def root_locus(sys, kvect=None, xlim=None, ylim=None,
               plotstr=None, plot=True, print_gain=None, grid=None, ax=None,
               initial_gain=None, **kwargs):

    """Root locus plot

    Calculate the root locus by finding the roots of 1+k*TF(s)
    where TF is self.num(s)/self.den(s) and each k is an element
    of kvect.

    Parameters
    ----------
    sys : LTI object
        Linear input/output systems (SISO only, for now).
    kvect : array_like, optional
        Gains to use in computing plot of closed-loop poles.
    xlim : tuple or list, optional
        Set limits of x axis, normally with tuple
        (see :doc:`matplotlib:api/axes_api`).
    ylim : tuple or list, optional
        Set limits of y axis, normally with tuple
        (see :doc:`matplotlib:api/axes_api`).
    plotstr : :func:`matplotlib.pyplot.plot` format string, optional
        plotting style specification
    plot : boolean, optional
        If True (default), plot root locus diagram.
    print_gain : bool
        If True (default), report mouse clicks when close to the root locus
        branches, calculate gain, damping and print.
    grid : bool
        If True plot omega-damping grid.  Default is False.
    ax : :class:`matplotlib.axes.Axes`
        Axes on which to create root locus plot
    initial_gain : float, optional
        Used by :func:`sisotool` to indicate initial gain.

    Returns
    -------
    roots : ndarray
        Closed-loop root locations, arranged in which each row corresponds
        to a gain in gains
    gains : ndarray
        Gains used.  Same as kvect keyword argument if provided.

    Notes
    -----
    The root_locus function calls matplotlib.pyplot.axis('equal'), which
    means that trying to reset the axis limits may not behave as expected.
    To change the axis limits, use matplotlib.pyplot.gca().axis('auto') and
    then set the axis limits to the desired values.

    """
    # Check to see if legacy 'Plot' keyword was used
    if 'Plot' in kwargs:
        warnings.warn("'Plot' keyword is deprecated in root_locus; "
                      "use 'plot'", FutureWarning)
        # Map 'Plot' keyword to 'plot' keyword
        plot = kwargs.pop('Plot')

    # Check to see if legacy 'PrintGain' keyword was used
    if 'PrintGain' in kwargs:
        warnings.warn("'PrintGain' keyword is deprecated in root_locus; "
                      "use 'print_gain'", FutureWarning)
        # Map 'PrintGain' keyword to 'print_gain' keyword
        print_gain = kwargs.pop('PrintGain')

    # Get parameter values
    plotstr = config._get_param('rlocus', 'plotstr', plotstr, _rlocus_defaults)
    grid = config._get_param('rlocus', 'grid', grid, _rlocus_defaults)
    print_gain = config._get_param(
        'rlocus', 'print_gain', print_gain, _rlocus_defaults)

    # Check for sisotool mode
    sisotool = kwargs.get('sisotool', False)

    # make sure siso. sisotool has different requirements
    if not sys.issiso() and not sisotool:
        raise ControlMIMONotImplemented(
            'sys must be single-input single-output (SISO)')

    sys_loop = sys[0,0]
    # Convert numerator and denominator to polynomials if they aren't
    (nump, denp) = _systopoly1d(sys_loop)

    # if discrete-time system and if xlim and ylim are not given,
    #  that we a view of the unit circle
    if xlim is None and isdtime(sys, strict=True):
        xlim = (-1.2, 1.2)
    if ylim is None and isdtime(sys, strict=True):
        xlim = (-1.3, 1.3)

    if kvect is None:
        kvect, root_array, xlim, ylim = _default_gains(nump, denp, xlim, ylim)
        recompute_on_zoom = True
    else:
        kvect = np.atleast_1d(kvect)
        root_array = _RLFindRoots(nump, denp, kvect)
        root_array = _RLSortRoots(root_array)
        recompute_on_zoom = False

    if sisotool:
        start_roots = _RLFindRoots(nump, denp, initial_gain)

    # Make sure there were no extraneous keywords
    if not sisotool and kwargs:
        raise TypeError("unrecognized keywords: ", str(kwargs))

    # Create the Plot
    if plot:
        if sisotool:
            fig = kwargs['fig']
            ax = fig.axes[1]
        else:
            if ax is None:
                ax = plt.gca()
            fig = ax.figure
            ax.set_title('Root Locus')

        if print_gain and not sisotool:
            fig.canvas.mpl_connect(
                'button_release_event',
                partial(_RLClickDispatcher, sys=sys, fig=fig,
                        ax_rlocus=fig.axes[0], plotstr=plotstr))
        elif sisotool:
            fig.axes[1].plot(
                [root.real for root in start_roots],
                [root.imag for root in start_roots],
                marker='s', markersize=6, zorder=20, color='k', label='gain_point')
            s = start_roots[0][0]
            if isdtime(sys, strict=True):
                zeta = -np.cos(np.angle(np.log(s)))
            else:
                zeta = -1 * s.real / abs(s)
            fig.suptitle(
                "Clicked at: %10.4g%+10.4gj  gain: %10.4g  damp: %10.4g" %
                (s.real, s.imag, initial_gain, zeta),
                fontsize=12 if int(mpl.__version__[0]) == 1 else 10)
            fig.canvas.mpl_connect(
                'button_release_event',
                partial(_RLClickDispatcher, sys=sys, fig=fig,
                        ax_rlocus=fig.axes[1], plotstr=plotstr,
                        sisotool=sisotool,
                        bode_plot_params=kwargs['bode_plot_params'],
                        tvect=kwargs['tvect']))


        if recompute_on_zoom:
            # update gains and roots when xlim/ylim change. Only then are
            # data on available. I.e., cannot combine with _RLClickDispatcher
            dpfun = partial(
                _RLZoomDispatcher, sys=sys, ax_rlocus=ax, plotstr=plotstr)
            # TODO: the next too lines seem to take a long time to execute
            # TODO: is there a way to speed them up?  (RMM, 6 Jun 2019)
            ax.callbacks.connect('xlim_changed', dpfun)
            ax.callbacks.connect('ylim_changed', dpfun)

        # plot open loop poles
        poles = array(denp.r)
        ax.plot(real(poles), imag(poles), 'x')

        # plot open loop zeros
        zeros = array(nump.r)
        if zeros.size > 0:
            ax.plot(real(zeros), imag(zeros), 'o')

        # Now plot the loci
        for index, col in enumerate(root_array.T):
            ax.plot(real(col), imag(col), plotstr, label='rootlocus')

        # Set up plot axes and labels
        ax.set_xlabel('Real')
        ax.set_ylabel('Imaginary')

        # Set up the limits for the plot
        # Note: need to do this before computing grid lines
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)

        # Draw the grid
        if grid:
            if isdtime(sys, strict=True):
                zgrid(ax=ax)
            else:
                _sgrid_func(fig=fig if sisotool else None)
        else:
            ax.axhline(0., linestyle=':', color='k', linewidth=.75, zorder=-20)
            ax.axvline(0., linestyle=':', color='k', linewidth=.75, zorder=-20)
            if isdtime(sys, strict=True):
                ax.add_patch(plt.Circle(
                    (0, 0), radius=1.0, linestyle=':', edgecolor='k',
                    linewidth=0.75, fill=False, zorder=-20))

    return root_array, kvect


def _default_gains(num, den, xlim, ylim, zoom_xlim=None, zoom_ylim=None):
    """Unsupervised gains calculation for root locus plot.

    References
    ----------
    Ogata, K. (2002). Modern control engineering (4th ed.). Upper
    Saddle River, NJ : New Delhi: Prentice Hall..

    """
    k_break, real_break = _break_points(num, den)
    kmax = _k_max(num, den, real_break, k_break)
    kvect = np.hstack((np.linspace(0, kmax, 50), np.real(k_break)))
    kvect.sort()

    root_array = _RLFindRoots(num, den, kvect)
    root_array = _RLSortRoots(root_array)
    open_loop_poles = den.roots
    open_loop_zeros = num.roots

    if open_loop_zeros.size != 0 and \
       open_loop_zeros.size < open_loop_poles.size:
        open_loop_zeros_xl = np.append(
            open_loop_zeros,
            np.ones(open_loop_poles.size - open_loop_zeros.size)
            * open_loop_zeros[-1])
        root_array_xl = np.append(root_array, open_loop_zeros_xl)
    else:
        root_array_xl = root_array
    singular_points = np.concatenate((num.roots, den.roots), axis=0)
    important_points = np.concatenate((singular_points, real_break), axis=0)
    important_points = np.concatenate((important_points, np.zeros(2)), axis=0)
    root_array_xl = np.append(root_array_xl, important_points)

    false_gain = float(den.coeffs[0]) / float(num.coeffs[0])
    if false_gain < 0 and not den.order > num.order:
        # TODO: make error message more understandable
        raise ValueError("Not implemented support for 0 degrees root locus "
                         "with equal order of numerator and denominator.")

    if xlim is None and false_gain > 0:
        x_tolerance = 0.05 * (np.max(np.real(root_array_xl))
                              - np.min(np.real(root_array_xl)))
        xlim = _ax_lim(root_array_xl)
    elif xlim is None and false_gain < 0:
        axmin = np.min(np.real(important_points)) \
            - (np.max(np.real(important_points))
               - np.min(np.real(important_points)))
        axmin = np.min(np.array([axmin, np.min(np.real(root_array_xl))]))
        axmax = np.max(np.real(important_points)) \
            + np.max(np.real(important_points)) \
            - np.min(np.real(important_points))
        axmax = np.max(np.array([axmax, np.max(np.real(root_array_xl))]))
        xlim = [axmin, axmax]
        x_tolerance = 0.05 * (axmax - axmin)
    else:
        x_tolerance = 0.05 * (xlim[1] - xlim[0])

    if ylim is None:
        y_tolerance = 0.05 * (np.max(np.imag(root_array_xl))
                              - np.min(np.imag(root_array_xl)))
        ylim = _ax_lim(root_array_xl * 1j)
    else:
        y_tolerance = 0.05 * (ylim[1] - ylim[0])

    # Figure out which points are spaced too far apart
    if x_tolerance == 0:
        # Root locus is on imaginary axis (rare), use just y distance
        tolerance = y_tolerance
    elif y_tolerance == 0:
        # Root locus is on imaginary axis (common), use just x distance
        tolerance = x_tolerance
    else:
        tolerance = np.min([x_tolerance, y_tolerance])
    indexes_too_far = _indexes_filt(root_array, tolerance, zoom_xlim, zoom_ylim)

    # Add more points into the root locus for points that are too far apart
    while len(indexes_too_far) > 0 and kvect.size < 5000:
        for counter, index in enumerate(indexes_too_far):
            index = index + counter*3
            new_gains = np.linspace(kvect[index], kvect[index + 1], 5)
            new_points = _RLFindRoots(num, den, new_gains[1:4])
            kvect = np.insert(kvect, index + 1, new_gains[1:4])
            root_array = np.insert(root_array, index + 1, new_points, axis=0)

        root_array = _RLSortRoots(root_array)
        indexes_too_far = _indexes_filt(root_array, tolerance, zoom_xlim, zoom_ylim)

    new_gains = kvect[-1] * np.hstack((np.logspace(0, 3, 4)))
    new_points = _RLFindRoots(num, den, new_gains[1:4])
    kvect = np.append(kvect, new_gains[1:4])
    root_array = np.concatenate((root_array, new_points), axis=0)
    root_array = _RLSortRoots(root_array)
    return kvect, root_array, xlim, ylim


def _indexes_filt(root_array, tolerance, zoom_xlim=None, zoom_ylim=None):
    """Calculate the distance between points and return the indexes.

    Filter the indexes so only the resolution of points within the xlim and
    ylim is improved when zoom is used.

    """
    distance_points = np.abs(np.diff(root_array, axis=0))
    indexes_too_far = list(np.unique(np.where(distance_points > tolerance)[0]))

    if zoom_xlim is not None and zoom_ylim is not None:
        x_tolerance_zoom = 0.05 * (zoom_xlim[1] - zoom_xlim[0])
        y_tolerance_zoom = 0.05 * (zoom_ylim[1] - zoom_ylim[0])
        tolerance_zoom = np.min([x_tolerance_zoom, y_tolerance_zoom])
        indexes_too_far_zoom = list(
            np.unique(np.where(distance_points > tolerance_zoom)[0]))
        indexes_too_far_filtered = []

        for index in indexes_too_far_zoom:
            for point in root_array[index]:
                if (zoom_xlim[0] <= point.real <= zoom_xlim[1]) and \
                   (zoom_ylim[0] <= point.imag <= zoom_ylim[1]):
                    indexes_too_far_filtered.append(index)
                    break

        # Check if zoom box is not overshot & insert points where neccessary
        if len(indexes_too_far_filtered) == 0 and len(root_array) < 500:
            limits = [zoom_xlim[0], zoom_xlim[1], zoom_ylim[0], zoom_ylim[1]]
            for index, limit in enumerate(limits):
                if index <= 1:
                    asign = np.sign(real(root_array)-limit)
                else:
                    asign = np.sign(imag(root_array) - limit)
                signchange = ((np.roll(asign, 1, axis=0)
                               - asign) != 0).astype(int)
                signchange[0] = np.zeros((len(root_array[0])))
                if len(np.where(signchange == 1)[0]) > 0:
                    indexes_too_far_filtered.append(
                        np.where(signchange == 1)[0][0]-1)

        if len(indexes_too_far_filtered) > 0:
            if indexes_too_far_filtered[0] != 0:
                indexes_too_far_filtered.insert(
                    0, indexes_too_far_filtered[0]-1)
            if not indexes_too_far_filtered[-1] + 1 >= len(root_array) - 2:
                indexes_too_far_filtered.append(
                    indexes_too_far_filtered[-1] + 1)

        indexes_too_far.extend(indexes_too_far_filtered)

    indexes_too_far = list(np.unique(indexes_too_far))
    indexes_too_far.sort()
    return indexes_too_far


def _break_points(num, den):
    """Extract break points over real axis and gains given these locations"""
    # type: (np.poly1d, np.poly1d) -> (np.array, np.array)
    dnum = num.deriv(m=1)
    dden = den.deriv(m=1)
    polynom = den * dnum - num * dden
    real_break_pts = polynom.r
    # don't care about infinite break points
    real_break_pts = real_break_pts[num(real_break_pts) != 0]
    k_break = -den(real_break_pts) / num(real_break_pts)
    idx = k_break >= 0   # only positives gains
    k_break = k_break[idx]
    real_break_pts = real_break_pts[idx]
    if len(k_break) == 0:
        k_break = [0]
        real_break_pts = den.roots
    return k_break, real_break_pts


def _ax_lim(root_array):
    """Utility to get the axis limits"""
    axmin = np.min(np.real(root_array))
    axmax = np.max(np.real(root_array))
    if axmax != axmin:
        deltax = (axmax - axmin) * 0.02
    else:
        deltax = np.max([1., axmax / 2])
    axlim = [axmin - deltax, axmax + deltax]
    return axlim


def _k_max(num, den, real_break_points, k_break_points):
    """"Calculate the maximum gain for the root locus shown in the figure."""
    asymp_number = den.order - num.order
    singular_points = np.concatenate((num.roots, den.roots), axis=0)
    important_points = np.concatenate(
        (singular_points, real_break_points), axis=0)
    false_gain = den.coeffs[0] / num.coeffs[0]

    if asymp_number > 0:
        asymp_center = (np.sum(den.roots) - np.sum(num.roots))/asymp_number
        distance_max = 4 * np.max(np.abs(important_points - asymp_center))
        asymp_angles = (2 * np.arange(0, asymp_number) - 1) \
            * np.pi / asymp_number
        if false_gain > 0:
            # farthest points over asymptotes
            farthest_points = asymp_center \
                + distance_max * np.exp(asymp_angles * 1j)
        else:
            asymp_angles = asymp_angles + np.pi
            # farthest points over asymptotes
            farthest_points = asymp_center \
                + distance_max * np.exp(asymp_angles * 1j)
        kmax_asymp = np.real(np.abs(den(farthest_points)
                                    / num(farthest_points)))
    else:
        kmax_asymp = np.abs([np.abs(den.coeffs[0])
                             / np.abs(num.coeffs[0]) * 3])

    kmax = np.max(np.concatenate((np.real(kmax_asymp),
                                  np.real(k_break_points)), axis=0))
    if np.abs(false_gain) > kmax:
        kmax = np.abs(false_gain)
    return kmax


def _systopoly1d(sys):
    """Extract numerator and denominator polynomails for a system"""
    # Allow inputs from the signal processing toolbox
    if (isinstance(sys, scipy.signal.lti)):
        nump = sys.num
        denp = sys.den

    else:
        # Convert to a transfer function, if needed
        sys = _convert_to_transfer_function(sys)

        # Make sure we have a SISO system
        if not sys.issiso():
            raise ControlMIMONotImplemented()

        # Start by extracting the numerator and denominator from system object
        nump = sys.num[0][0]
        denp = sys.den[0][0]

    # Check to see if num, den are already polynomials; otherwise convert
    if (not isinstance(nump, poly1d)):
        nump = poly1d(nump)

    if (not isinstance(denp, poly1d)):
        denp = poly1d(denp)

    return (nump, denp)


def _RLFindRoots(nump, denp, kvect):
    """Find the roots for the root locus."""
    # Convert numerator and denominator to polynomials if they aren't
    roots = []
    for k in np.atleast_1d(kvect):
        curpoly = denp + k * nump
        curroots = curpoly.r
        if len(curroots) < denp.order:
            # if I have fewer poles than open loop, it is because i have
            # one at infinity
            curroots = np.append(curroots, np.inf)

        curroots.sort()
        roots.append(curroots)

    return row_stack(roots)


def _RLSortRoots(roots):
    """Sort the roots from _RLFindRoots, so that the root
    locus doesn't show weird pseudo-branches as roots jump from
    one branch to another."""

    sorted = zeros_like(roots)
    for n, row in enumerate(roots):
        if n == 0:
            sorted[n, :] = row
        else:
            # sort the current row by finding the element with the
            # smallest absolute distance to each root in the
            # previous row
            available = list(range(len(prevrow)))
            for elem in row:
                evect = elem - prevrow[available]
                ind1 = abs(evect).argmin()
                ind = available.pop(ind1)
                sorted[n, ind] = elem
        prevrow = sorted[n, :]
    return sorted


def _RLZoomDispatcher(event, sys, ax_rlocus, plotstr):
    """Rootlocus plot zoom dispatcher"""
    sys_loop = sys[0,0]
    nump, denp = _systopoly1d(sys_loop)
    xlim, ylim = ax_rlocus.get_xlim(), ax_rlocus.get_ylim()

    kvect, root_array, xlim, ylim = _default_gains(
        nump, denp, xlim=None, ylim=None, zoom_xlim=xlim, zoom_ylim=ylim)
    _removeLine('rootlocus', ax_rlocus)

    for i, col in enumerate(root_array.T):
        ax_rlocus.plot(real(col), imag(col), plotstr, label='rootlocus',
                       scalex=False, scaley=False)


def _RLClickDispatcher(event, sys, fig, ax_rlocus, plotstr, sisotool=False,
                       bode_plot_params=None, tvect=None):
    """Rootlocus plot click dispatcher"""

    # Zoom is handled by specialized callback above, only do gain plot
    if event.inaxes == ax_rlocus.axes and \
       plt.get_current_fig_manager().toolbar.mode not in \
       {'zoom rect', 'pan/zoom'}:

        # if a point is clicked on the rootlocus plot visually emphasize it
        K = _RLFeedbackClicksPoint(event, sys, fig, ax_rlocus, sisotool)
        if sisotool and K is not None:
            _SisotoolUpdate(sys, fig, K, bode_plot_params, tvect)

    # Update the canvas
    fig.canvas.draw()


def _RLFeedbackClicksPoint(event, sys, fig, ax_rlocus, sisotool=False):
    """Display root-locus gain feedback point for clicks on root-locus plot"""
    sys_loop = sys[0,0]
    (nump, denp) = _systopoly1d(sys_loop)

    xlim = ax_rlocus.get_xlim()
    ylim = ax_rlocus.get_ylim()
    x_tolerance = 0.1 * abs((xlim[1] - xlim[0]))
    y_tolerance = 0.1 * abs((ylim[1] - ylim[0]))
    gain_tolerance = np.mean([x_tolerance, y_tolerance])*0.1

    # Catch type error when event click is in the figure but not in an axis
    try:
        s = complex(event.xdata, event.ydata)
        K = -1. / sys_loop(s)
        K_xlim = -1. / sys_loop(
            complex(event.xdata + 0.05 * abs(xlim[1] - xlim[0]), event.ydata))
        K_ylim = -1. / sys_loop(
            complex(event.xdata, event.ydata + 0.05 * abs(ylim[1] - ylim[0])))

    except TypeError:
        K = float('inf')
        K_xlim = float('inf')
        K_ylim = float('inf')

    gain_tolerance += 0.1 * max([abs(K_ylim.imag/K_ylim.real),
                                 abs(K_xlim.imag/K_xlim.real)])

    if abs(K.real) > 1e-8 and abs(K.imag / K.real) < gain_tolerance and \
       event.inaxes == ax_rlocus.axes and K.real > 0.:

        if isdtime(sys, strict=True):
            zeta = -np.cos(np.angle(np.log(s)))
        else:
            zeta = -1 * s.real / abs(s)

        # Display the parameters in the output window and figure
        print("Clicked at %10.4g%+10.4gj gain %10.4g damp %10.4g" %
              (s.real, s.imag, K.real, zeta))
        fig.suptitle(
            "Clicked at: %10.4g%+10.4gj  gain: %10.4g  damp: %10.4g" %
            (s.real, s.imag, K.real, zeta),
            fontsize=12 if int(mpl.__version__[0]) == 1 else 10)

        # Remove the previous line
        _removeLine(label='gain_point', ax=ax_rlocus)

        # Visualise clicked point, display all roots for sisotool mode
        if sisotool:
            root_array = _RLFindRoots(nump, denp, K.real)
            ax_rlocus.plot(
                [root.real for root in root_array],
                [root.imag for root in root_array],
                marker='s', markersize=6, zorder=20, label='gain_point', color='k')
        else:
            ax_rlocus.plot(s.real, s.imag, 'k.', marker='s', markersize=8,
                           zorder=20, label='gain_point')

        return K.real


def _removeLine(label, ax):
    """Remove a line from the ax when a label is specified"""
    for line in reversed(ax.lines):
        if line.get_label() == label:
            line.remove()
            del line


def _sgrid_func(fig=None, zeta=None, wn=None):
    if fig is None:
        fig = plt.gcf()
        ax = fig.gca()
    else:
        ax = fig.axes[1]

    # Get locator function for x-axis, y-axis tick marks
    xlocator = ax.get_xaxis().get_major_locator()
    ylocator = ax.get_yaxis().get_major_locator()

    # Decide on the location for the labels (?)
    ylim = ax.get_ylim()
    ytext_pos_lim = ylim[1] - (ylim[1] - ylim[0]) * 0.03
    xlim = ax.get_xlim()
    xtext_pos_lim = xlim[0] + (xlim[1] - xlim[0]) * 0.0

    # Create a list of damping ratios, if needed
    if zeta is None:
        zeta = _default_zetas(xlim, ylim)

    # Figure out the angles for the different damping ratios
    angles = []
    for z in zeta:
        if (z >= 1e-4) and (z <= 1):
            angles.append(np.pi/2 + np.arcsin(z))
        else:
            zeta.remove(z)
    y_over_x = np.tan(angles)

    # zeta-constant lines
    for index, yp in enumerate(y_over_x):
        ax.plot([0, xlocator()[0]], [0, yp * xlocator()[0]], color='gray',
                linestyle='dashed', linewidth=0.5)
        ax.plot([0, xlocator()[0]], [0, -yp * xlocator()[0]], color='gray',
                linestyle='dashed', linewidth=0.5)
        an = "%.2f" % zeta[index]
        if yp < 0:
            xtext_pos = 1/yp * ylim[1]
            ytext_pos = yp * xtext_pos_lim
            if np.abs(xtext_pos) > np.abs(xtext_pos_lim):
                xtext_pos = xtext_pos_lim
            else:
                ytext_pos = ytext_pos_lim
            ax.annotate(an, textcoords='data', xy=[xtext_pos, ytext_pos],
                        fontsize=8)
    ax.plot([0, 0], [ylim[0], ylim[1]],
            color='gray', linestyle='dashed', linewidth=0.5)

    # omega-constant lines
    angles = np.linspace(-90, 90, 20) * np.pi/180
    if wn is None:
        wn = _default_wn(xlocator(), ylocator())

    for om in wn:
        if om < 0:
            # Generate the lines for natural frequency curves
            yp = np.sin(angles) * np.abs(om)
            xp = -np.cos(angles) * np.abs(om)

            # Plot the natural frequency contours
            ax.plot(xp, yp, color='gray', linestyle='dashed', linewidth=0.5)

            # Annotate the natural frequencies by listing on x-axis
            # Note: need to filter values for proper plotting in Jupyter
            if (om > xlim[0]):
                an = "%.2f" % -om
                ax.annotate(an, textcoords='data', xy=[om, 0], fontsize=8)


def _default_zetas(xlim, ylim):
    """Return default list of damping coefficients

    This function computes a list of damping coefficients based on the limits
    of the graph.  A set of 4 damping coefficients are computed for the x-axis
    and a set of three damping coefficients are computed for the y-axis
    (corresponding to the normal 4:3 plot aspect ratio in `matplotlib`?).

    Parameters
    ----------
    xlim : array_like
        List of x-axis limits [min, max]
    ylim : array_like
        List of y-axis limits [min, max]

    Returns
    -------
    zeta : list
        List of default damping coefficients for the plot

    """
    # Damping coefficient lines that intersect the x-axis
    sep1 = -xlim[0] / 4
    ang1 = [np.arctan((sep1*i)/ylim[1]) for i in np.arange(1, 4, 1)]

    # Damping coefficient lines that intersection the y-axis
    sep2 = ylim[1] / 3
    ang2 = [np.arctan(-xlim[0]/(ylim[1]-sep2*i)) for i in np.arange(1, 3, 1)]

    # Put the lines together and add one at -pi/2 (negative real axis)
    angles = np.concatenate((ang1, ang2))
    angles = np.insert(angles, len(angles), np.pi/2)

    # Return the damping coefficients corresponding to these angles
    zeta = np.sin(angles)
    return zeta.tolist()


def _default_wn(xloc, yloc, max_lines=7):
    """Return default wn for root locus plot

    This function computes a list of natural frequencies based on the grid
    parameters of the graph.

    Parameters
    ----------
    xloc : array_like
        List of x-axis tick values
    ylim : array_like
        List of y-axis limits [min, max]
    max_lines : int, optional
        Maximum number of frequencies to generate (default = 7)

    Returns
    -------
    wn : list
        List of default natural frequencies for the plot

    """
    sep = xloc[1]-xloc[0]       # separation between x-ticks

    # Decide whether to use the x or y axis for determining wn
    if yloc[-1] / sep > max_lines*10:
        # y-axis scale >> x-axis scale
        wn = yloc                   # one frequency per y-axis tick mark
    else:
        wn = xloc                   # one frequency per x-axis tick mark

        # Insert additional frequencies to span the y-axis
        while np.abs(wn[0]) < yloc[-1]:
            wn = np.insert(wn, 0, wn[0]-sep)

    # If there are too many values, cut them in half
    while len(wn) > max_lines:
        wn = wn[0:-1:2]

    return wn


rlocus = root_locus



# basis.py - BasisFamily class
# RMM, 10 Nov 2012
#
# The BasisFamily class is used to specify a set of basis functions for
# implementing differential flatness computations.
#
# Copyright (c) 2012 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import numpy as np


# Basis family class (for use as a base class)
class BasisFamily:
    """Base class for implementing basis functions for flat systems.

    A BasisFamily object is used to construct trajectories for a flat system.
    The class must implement a single function that computes the jth
    derivative of the ith basis function at a time t:

      :math:`z_i^{(q)}(t)` = basis.eval_deriv(self, i, j, t)

    A basis set can either consist of a single variable that is used for
    each flat output (nvars = None) or a different variable for different
    flat outputs (nvars > 0).

    Attributes
    ----------
    N : int
        Order of the basis set.

    """
    def __init__(self, N):
        """Create a basis family of order N."""
        self.N = N                    # save number of basis functions
        self.nvars = None             # default number of variables
        self.coef_offset = [0]        # coefficient offset for each variable
        self.coef_length = [N]        # coefficient length for each variable

    def __repr__(self):
        return f'<{self.__class__.__name__}: nvars={self.nvars}, ' + \
            f'N={self.N}>'

    def __call__(self, i, t, var=None):
        """Evaluate the ith basis function at a point in time"""
        return self.eval_deriv(i, 0, t, var=var)

    def var_ncoefs(self, var):
        """Get the number of coefficients for a variable"""
        return self.N if self.nvars is None else self.coef_length[var]

    def eval(self, coeffs, tlist, var=None):
        """Compute function values given the coefficients and time points."""
        if self.nvars is None and var != None:
            raise SystemError("multi-variable call to a scalar basis")

        elif self.nvars is None:
            # Single variable basis
            return [
                sum([coeffs[i] * self(i, t) for i in range(self.N)])
                for t in tlist]

        elif var is None:
            # Multi-variable basis with single list of coefficients
            values = np.empty((self.nvars, tlist.size))
            offset = 0
            for j in range(self.nvars):
                coef_len = self.var_ncoefs(j)
                values[j] = np.array([
                    sum([coeffs[offset + i] * self(i, t, var=j)
                         for i in range(coef_len)])
                    for t in tlist])
                offset += coef_len
            return values

        else:
            return np.array([
                sum([coeffs[i] * self(i, t, var=var)
                     for i in range(self.var_ncoefs(var))])
                for t in tlist])

    def eval_deriv(self, i, j, t, var=None):
        """Evaluate the kth derivative of the ith basis function at time t."""
        raise NotImplementedError("Internal error; improper basis functions")



# systraj.py - SystemTrajectory class
# RMM, 10 November 2012
#
# The SystemTrajetory class is used to store a feasible trajectory for
# the state and input of a (nonlinear) control system.
#
# Copyright (c) 2012 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import numpy as np
from ..timeresp import TimeResponseData

class SystemTrajectory:
    """Class representing a system trajectory.

    The `SystemTrajectory` class is used to represent the
    trajectory of a (differentially flat) system.  Used by the
    :func:`~control.trajsys.point_to_point` function to return a trajectory.

    Parameters
    ----------
    sys : FlatSystem
        Flat system object associated with this trajectory.
    basis : BasisFamily
        Family of basis vectors to use to represent the trajectory.
    coeffs : list of 1D arrays, optional
        For each flat output, define the coefficients of the basis
        functions used to represent the trajectory.  Defaults to an empty
        list.
    flaglen : list of ints, optional
        For each flat output, the number of derivatives of the flat
        output used to define the trajectory.  Defaults to an empty
        list.

    """

    def __init__(self, sys, basis, coeffs=[], flaglen=[], params=None):
        """Initilize a system trajectory object."""
        self.nstates = sys.nstates
        self.ninputs = sys.ninputs
        self.system = sys
        self.basis = basis
        self.coeffs = list(coeffs)
        self.flaglen = list(flaglen)
        self.params = sys.params if params is None else params

    # Evaluate the trajectory over a list of time points
    def eval(self, tlist):
        """Return the state and input for a trajectory at a list of times.

        Evaluate the trajectory at a list of time points, returning the state
        and input vectors for the trajectory:

            x, u = traj.eval(tlist)

        Parameters
        ----------
        tlist : 1D array
            List of times to evaluate the trajectory.

        Returns
        -------
        x : 2D array
            For each state, the values of the state at the given times.
        u : 2D array
            For each input, the values of the input at the given times.

        """
        # Allocate space for the outputs
        xd = np.zeros((self.nstates, len(tlist)))
        ud = np.zeros((self.ninputs, len(tlist)))

        # Go through each time point and compute xd and ud via flat variables
        # TODO: make this more pythonic
        for tind, t in enumerate(tlist):
            zflag = []
            for i in range(self.ninputs):
                flag_len = self.flaglen[i]
                zflag.append(np.zeros(flag_len))
                for j in range(self.basis.var_ncoefs(i)):
                    for k in range(flag_len):
                        #! TODO: rewrite eval_deriv to take in time vector
                        zflag[i][k] += self.coeffs[i][j] * \
                            self.basis.eval_deriv(j, k, t, var=i)

            # Now copy the states and inputs
            # TODO: revisit order of list arguments
            xd[:,tind], ud[:,tind] = \
                self.system.reverse(zflag, self.params)

        return xd, ud

    # Return the system trajectory as a TimeResponseData object
    def response(self, tlist, transpose=False, return_x=False, squeeze=None):
        """Return the trajectory of a system as a TimeResponseData object

        Evaluate the trajectory at a list of time points, returning the state
        and input vectors for the trajectory:

            response = traj.response(tlist)
            time, yd, ud = response.time, response.outputs, response.inputs

        Parameters
        ----------
        tlist : 1D array
            List of times to evaluate the trajectory.

        transpose : bool, optional
            If True, transpose all input and output arrays (for backward
            compatibility with MATLAB and :func:`scipy.signal.lsim`).
            Default value is False.

        return_x : bool, optional
            If True, return the state vector when assigning to a tuple
            (default = False).  See :func:`forced_response` for more details.

        squeeze : bool, optional
            By default, if a system is single-input, single-output (SISO) then
            the output response is returned as a 1D array (indexed by time).
            If squeeze=True, remove single-dimensional entries from the shape
            of the output even if the system is not SISO. If squeeze=False,
            keep the output as a 3D array (indexed by the output, input, and
            time) even if the system is SISO. The default value can be set
            using config.defaults['control.squeeze_time_response'].

        Returns
        -------
        results : TimeResponseData
            Time response represented as a :class:`TimeResponseData` object
            containing the following properties:

            * time (array): Time values of the output.

            * outputs (array): Response of the system.  If the system is SISO
              and squeeze is not True, the array is 1D (indexed by time).  If
              the system is not SISO or ``squeeze`` is False, the array is 3D
              (indexed by the output, trace, and time).

            * states (array): Time evolution of the state vector, represented
              as either a 2D array indexed by state and time (if SISO) or a 3D
              array indexed by state, trace, and time.  Not affected by
              ``squeeze``.

            * inputs (array): Input(s) to the system, indexed in the same
              manner as ``outputs``.

            The return value of the system can also be accessed by assigning
            the function to a tuple of length 2 (time, output) or of length 3
            (time, output, state) if ``return_x`` is ``True``.

        """
        # Compute the state and input response using the eval function
        sys = self.system
        xout, uout = self.eval(tlist)
        yout = np.array([
            sys.output(tlist[i], xout[:, i], uout[:, i])
            for i in range(len(tlist))]).transpose()

        return TimeResponseData(
            tlist, yout, xout, uout, issiso=sys.issiso(),
            input_labels=sys.input_labels, output_labels=sys.output_labels,
            state_labels=sys.state_labels,
            transpose=transpose, return_x=return_x, squeeze=squeeze)



# steering-gainsched.py - gain scheduled control for vehicle steering
# RMM, 8 May 2019
#
# This file works through Example 1.1 in the "Optimization-Based Control"
# course notes by Richard Murray (avaliable at http://fbsbook.org, in the
# optimization-based control supplement).  It is intended to demonstrate the
# functionality for nonlinear input/output systems in the python-control
# package.

import numpy as np
import control as ct
from cmath import sqrt
import matplotlib.pyplot as plt

#
# Vehicle steering dynamics
#
# The vehicle dynamics are given by a simple bicycle model.  We take the state
# of the system as (x, y, theta) where (x, y) is the position of the vehicle
# in the plane and theta is the angle of the vehicle with respect to
# horizontal.  The vehicle input is given by (v, phi) where v is the forward
# velocity of the vehicle and phi is the angle of the steering wheel.  The
# model includes saturation of the vehicle steering angle.
#
# System state: x, y, theta
# System input: v, phi
# System output: x, y
# System parameters: wheelbase, maxsteer
#
def vehicle_update(t, x, u, params):
    # Get the parameters for the model
    l = params.get('wheelbase', 3.)         # vehicle wheelbase
    phimax = params.get('maxsteer', 0.5)    # max steering angle (rad)

    # Saturate the steering input
    phi = np.clip(u[1], -phimax, phimax)

    # Return the derivative of the state
    return np.array([
        np.cos(x[2]) * u[0],            # xdot = cos(theta) v
        np.sin(x[2]) * u[0],            # ydot = sin(theta) v
        (u[0] / l) * np.tan(phi)        # thdot = v/l tan(phi)
    ])

def vehicle_output(t, x, u, params):
    return x                            # return x, y, theta (full state)

# Define the vehicle steering dynamics as an input/output system
vehicle = ct.NonlinearIOSystem(
    vehicle_update, vehicle_output, states=3, name='vehicle',
    inputs=('v', 'phi'),
    outputs=('x', 'y', 'theta'))

#
# Gain scheduled controller
#
# For this system we use a simple schedule on the forward vehicle velocity and
# place the poles of the system at fixed values.  The controller takes the
# current vehicle position and orientation plus the velocity velocity as
# inputs, and returns the velocity and steering commands.
#
# System state: none
# System input: ex, ey, etheta, vd, phid
# System output: v, phi
# System parameters: longpole, latpole1, latpole2
#
def control_output(t, x, u, params):
    # Get the controller parameters
    longpole = params.get('longpole', -2.)
    latpole1 = params.get('latpole1', -1/2 + sqrt(-7)/2)
    latpole2 = params.get('latpole2', -1/2 - sqrt(-7)/2)
    l = params.get('wheelbase', 3)

    # Extract the system inputs
    ex, ey, etheta, vd, phid = u

    # Determine the controller gains
    alpha1 = -np.real(latpole1 + latpole2)
    alpha2 = np.real(latpole1 * latpole2)

    # Compute and return the control law
    v = -longpole * ex          # Note: no feedfwd (to make plot interesting)
    if vd != 0:
        phi = phid + (alpha1 * l) / vd * ey + (alpha2 * l) / vd * etheta
    else:
        # We aren't moving, so don't turn the steering wheel
        phi = phid

    return  np.array([v, phi])

# Define the controller as an input/output system
controller = ct.NonlinearIOSystem(
    None, control_output, name='controller',        # static system
    inputs=('ex', 'ey', 'etheta', 'vd', 'phid'),    # system inputs
    outputs=('v', 'phi')                            # system outputs
)

#
# Reference trajectory subsystem
#
# The reference trajectory block generates a simple trajectory for the system
# given the desired speed (vref) and lateral position (yref).  The trajectory
# consists of a straight line of the form (vref * t, yref, 0) with nominal
# input (vref, 0).
#
# System state: none
# System input: vref, yref
# System output: xd, yd, thetad, vd, phid
# System parameters: none
#
def trajgen_output(t, x, u, params):
    vref, yref = u
    return np.array([vref * t, yref, 0, vref, 0])

# Define the trajectory generator as an input/output system
trajgen = ct.NonlinearIOSystem(
    None, trajgen_output, name='trajgen',
    inputs=('vref', 'yref'),
    outputs=('xd', 'yd', 'thetad', 'vd', 'phid'))

#
# System construction
#
# The input to the full closed loop system is the desired lateral position and
# the desired forward velocity.  The output for the system is taken as the
# full vehicle state plus the velocity of the vehicle.  The following diagram
# summarizes the interconnections:
#
#                        +---------+       +---------------> v
#                        |         |       |
# [ yref ]               |         v       |
# [      ] ---> trajgen -+-+-> controller -+-> vehicle -+-> [x, y, theta]
# [ vref ]                 ^                            |
#                          |                            |
#                          +----------- [-1] -----------+
#
# We construct the system using the InterconnectedSystem constructor and using
# signal labels to keep track of everything.

steering = ct.interconnect(
    # List of subsystems
    (trajgen, controller, vehicle), name='steering',

    # Interconnections between  subsystems
    connections=(
        ['controller.ex', 'trajgen.xd', '-vehicle.x'],
        ['controller.ey', 'trajgen.yd', '-vehicle.y'],
        ['controller.etheta', 'trajgen.thetad', '-vehicle.theta'],
        ['controller.vd', 'trajgen.vd'],
        ['controller.phid', 'trajgen.phid'],
        ['vehicle.v', 'controller.v'],
        ['vehicle.phi', 'controller.phi']
    ),

    # System inputs
    inplist=['trajgen.vref', 'trajgen.yref'],
    inputs=['yref', 'vref'],

    #  System outputs
    outlist=['vehicle.x', 'vehicle.y', 'vehicle.theta', 'controller.v',
             'controller.phi'],
    outputs=['x', 'y', 'theta', 'v', 'phi']
)

# Set up the simulation conditions
yref = 1
T = np.linspace(0, 5, 100)

# Set up a figure for plotting the results
plt.figure();

# Plot the reference trajectory for the y position
plt.plot([0, 5], [yref, yref], 'k--')

# Find the signals we want to plot
y_index = steering.find_output('y')
v_index = steering.find_output('v')

# Do an iteration through different speeds
for vref in [8, 10, 12]:
    # Simulate the closed loop controller response
    tout, yout = ct.input_output_response(
        steering, T, [vref * np.ones(len(T)), yref * np.ones(len(T))])

    # Plot the reference speed
    plt.plot([0, 5], [vref, vref], 'k--')

    # Plot the system output
    y_line, = plt.plot(tout, yout[y_index], 'r')  # lateral position
    v_line, = plt.plot(tout, yout[v_index], 'b')  # vehicle velocity

# Add axis labels
plt.xlabel('Time (s)')
plt.ylabel('x vel (m/s), y pos (m)')
plt.legend((v_line, y_line), ('v', 'y'), loc='center right', frameon=False)

#
# Alternative formulation, using create_statefbk_iosystem()
#
# A different way to implement gain scheduling is to use the gain scheduling
# functionality built into the create_statefbk_iosystem() function, where we
# pass a table of gains instead of a single gain.  To generate a more
# interesting plot, we scale the feedforward input to generate some error.
#

import itertools
from math import pi

# Define the points for the scheduling variables
speeds = [1, 10, 20]
angles = np.linspace(-pi, pi, 4)
points = list(itertools.product(speeds, angles))

# Create controllers at each scheduling point
Q = np.diag([1, 1, 1])
R = np.diag([0.1, 0.1])
gains = [np.array(ct.lqr(vehicle.linearize(
    [0, 0, angle], [speed, 0]), Q, R)[0]) for speed, angle in points]

# Create the gain scheduled system
controller, _ = ct.create_statefbk_iosystem(
    vehicle, (gains, points), name='controller', ud_labels=['vd', 'phid'],
    gainsched_indices=['vd', 'theta'], gainsched_method='linear')

# Connect everything together (note that controller inputs are different
steering = ct.interconnect(
    # List of subsystems
    (trajgen, controller, vehicle), name='steering',

    # Interconnections between  subsystems
    connections=(
        ['controller.xd[0]', 'trajgen.xd'],
        ['controller.xd[1]', 'trajgen.yd'],
        ['controller.xd[2]', 'trajgen.thetad'],
        ['controller.x', 'vehicle.x'],
        ['controller.y', 'vehicle.y'],
        ['controller.theta', 'vehicle.theta'],
        ['controller.vd', ('trajgen', 'vd', 0.2)],      # create error
        ['controller.phid', 'trajgen.phid'],
        ['vehicle.v', 'controller.v'],
        ['vehicle.phi', 'controller.phi']
    ),

    # System inputs
    inplist=['trajgen.vref', 'trajgen.yref'],
    inputs=['yref', 'vref'],

    #  System outputs
    outlist=['vehicle.x', 'vehicle.y', 'vehicle.theta', 'controller.v',
             'controller.phi'],
    outputs=['x', 'y', 'theta', 'v', 'phi']
)

# Plot the results to compare to the previous case
plt.figure();

# Plot the reference trajectory for the y position
plt.plot([0, 5], [yref, yref], 'k--')

# Find the signals we want to plot
y_index = steering.find_output('y')
v_index = steering.find_output('v')

# Do an iteration through different speeds
for vref in [8, 10, 12]:
    # Simulate the closed loop controller response
    tout, yout = ct.input_output_response(
        steering, T, [vref * np.ones(len(T)), yref * np.ones(len(T))])

    # Plot the reference speed
    plt.plot([0, 5], [vref, vref], 'k--')

    # Plot the system output
    y_line, = plt.plot(tout, yout[y_index], 'r')  # lateral position
    v_line, = plt.plot(tout, yout[v_index], 'b')  # vehicle velocity

# Add axis labels
plt.xlabel('Time (s)')
plt.ylabel('x vel (m/s), y pos (m)')
plt.legend((v_line, y_line), ('v', 'y'), loc='center right', frameon=False)



# pvtol.py - (planar) vertical takeoff and landing system model
# RMM, 19 Jan 2022
#
# This file contains a model and utility function for a (planar)
# vertical takeoff and landing system, as described in FBS2e and OBC.
# This system is approximately differentially flat and the flat system
# mappings are included.
#

import numpy as np
import matplotlib.pyplot as plt
import control as ct
import control.flatsys as fs
from math import sin, cos
from warnings import warn

# PVTOL dynamics
def pvtol_update(t, x, u, params):
    # Get the parameter values
    m = params.get('m', 4.)             # mass of aircraft
    J = params.get('J', 0.0475)         # inertia around pitch axis
    r = params.get('r', 0.25)           # distance to center of force
    g = params.get('g', 9.8)            # gravitational constant
    c = params.get('c', 0.05)           # damping factor (estimated)

    # Get the inputs and states
    x, y, theta, xdot, ydot, thetadot = x
    F1, F2 = u

    # Constrain the inputs
    F2 = np.clip(F2, 0, 1.5 * m * g)
    F1 = np.clip(F1, -0.1 * F2, 0.1 * F2)

    # Dynamics
    xddot = (F1 * cos(theta) - F2 * sin(theta) - c * xdot) / m
    yddot = (F1 * sin(theta) + F2 * cos(theta) - m * g - c * ydot) / m
    thddot = (r * F1) / J

    return np.array([xdot, ydot, thetadot, xddot, yddot, thddot])

def pvtol_output(t, x, u, params):
    return x

# PVTOL flat system mappings
def pvtol_flat_forward(states, inputs, params={}):
    # Get the parameter values
    m = params.get('m', 4.)             # mass of aircraft
    J = params.get('J', 0.0475)         # inertia around pitch axis
    r = params.get('r', 0.25)           # distance to center of force
    g = params.get('g', 9.8)            # gravitational constant
    c = params.get('c', 0.05)           # damping factor (estimated)

    # Make sure that c is zero
    if c != 0:
        warn("System is only approximately flat (c != 0)")

    # Create a list of arrays to store the flat output and its derivatives
    zflag = [np.zeros(5), np.zeros(5)]

    # Store states and inputs in variables to make things easier to read
    x, y, theta, xdot, ydot, thdot = states
    F1, F2 = inputs

    # Use equations of motion for higher derivates
    x1ddot = (F1 * cos(theta) - F2 * sin(theta)) / m
    x2ddot = (F1 * sin(theta) + F2 * cos(theta) - m * g) / m
    thddot = (r * F1) / J

    # Flat output is a point above the vertical axis
    zflag[0][0] = x - (J / (m * r)) * sin(theta)
    zflag[1][0] = y + (J / (m * r)) * cos(theta)

    zflag[0][1] = xdot - (J / (m * r)) * cos(theta) * thdot
    zflag[1][1] = ydot - (J / (m * r)) * sin(theta) * thdot

    zflag[0][2] = (F1 * cos(theta) - F2 * sin(theta)) / m \
        + (J / (m * r)) * sin(theta) * thdot**2 \
        - (J / (m * r)) * cos(theta) * thddot
    zflag[1][2] = (F1 * sin(theta) + F2 * cos(theta) - m * g) / m \
        - (J / (m * r)) * cos(theta) * thdot**2 \
        - (J / (m * r)) * sin(theta) * thddot

    # For the third derivative, assume F1, F2 are constant (also thddot)
    zflag[0][3] = (-F1 * sin(theta) - F2 * cos(theta)) * (thdot / m) \
        + (J / (m * r)) * cos(theta) * thdot**3 \
        + 3 * (J / (m * r)) * sin(theta) * thdot * thddot
    zflag[1][3] = (F1 * cos(theta) - F2 * sin(theta)) * (thdot / m) \
        + (J / (m * r)) * sin(theta) * thdot**3 \
        - 3 * (J / (m * r)) * cos(theta) * thdot * thddot

    # For the fourth derivative, assume F1, F2 are constant (also thddot)
    zflag[0][4] = (-F1 * sin(theta) - F2 * cos(theta)) * (thddot / m) \
        + (-F1 * cos(theta) + F2 * sin(theta)) * (thdot**2 / m) \
        + 6 * (J / (m * r)) * cos(theta) * thdot**2 * thddot \
        + 3 * (J / (m * r)) * sin(theta) * thddot**2 \
        - (J / (m * r)) * sin(theta) * thdot**4
    zflag[1][4] = (F1 * cos(theta) - F2 * sin(theta)) * (thddot / m) \
        + (-F1 * sin(theta) - F2 * cos(theta)) * (thdot**2 / m) \
        - 6 * (J / (m * r)) * sin(theta) * thdot**2 * thddot \
        - 3 * (J / (m * r)) * cos(theta) * thddot**2 \
        + (J / (m * r)) * cos(theta) * thdot**4

    return zflag

def pvtol_flat_reverse(zflag, params={}):
    # Get the parameter values
    m = params.get('m', 4.)             # mass of aircraft
    J = params.get('J', 0.0475)         # inertia around pitch axis
    r = params.get('r', 0.25)           # distance to center of force
    g = params.get('g', 9.8)            # gravitational constant
    c = params.get('c', 0.05)           # damping factor (estimated)

    # Create a vector to store the state and inputs
    x = np.zeros(6)
    u = np.zeros(2)

    # Given the flat variables, solve for the state
    theta = np.arctan2(-zflag[0][2],  zflag[1][2] + g)
    x = zflag[0][0] + (J / (m * r)) * sin(theta)
    y = zflag[1][0] - (J / (m * r)) * cos(theta)

    # Solve for thdot using next derivative
    thdot = (zflag[0][3] * cos(theta) + zflag[1][3] * sin(theta)) \
        / (zflag[0][2] * sin(theta) - (zflag[1][2] + g) * cos(theta))

    # xdot and ydot can be found from first derivative of flat outputs
    xdot = zflag[0][1] + (J / (m * r)) * cos(theta) * thdot
    ydot = zflag[1][1] + (J / (m * r)) * sin(theta) * thdot

    # Solve for the input forces
    F2 = m * ((zflag[1][2] + g) * cos(theta) - zflag[0][2] * sin(theta)
              + (J / (m * r)) * thdot**2)
    F1 = (J / r) * \
        (zflag[0][4] * cos(theta) + zflag[1][4] * sin(theta)
#         - 2 * (zflag[0][3] * sin(theta) - zflag[1][3] * cos(theta)) * thdot \
         - 2 * zflag[0][3] * sin(theta) * thdot \
         + 2 * zflag[1][3] * cos(theta) * thdot \
#         - (zflag[0][2] * cos(theta)
#            + (zflag[1][2] + g) * sin(theta)) * thdot**2) \
         - zflag[0][2] * cos(theta) * thdot**2
         - (zflag[1][2] + g) * sin(theta) * thdot**2) \
        / (zflag[0][2] * sin(theta) - (zflag[1][2] + g) * cos(theta))

    return np.array([x, y, theta, xdot, ydot, thdot]), np.array([F1, F2])

pvtol = fs.FlatSystem(
    pvtol_flat_forward, pvtol_flat_reverse, name='pvtol',
    updfcn=pvtol_update, outfcn=pvtol_output,
    states = [f'x{i}' for i in range(6)],
    inputs = ['F1', 'F2'],
    outputs = [f'x{i}' for i in range(6)],
    params = {
        'm': 4.,                # mass of aircraft
        'J': 0.0475,            # inertia around pitch axis
        'r': 0.25,              # distance to center of force
        'g': 9.8,               # gravitational constant
        'c': 0.05,              # damping factor (estimated)
    }
)

#
# PVTOL dynamics with wind
# 

def windy_update(t, x, u, params):
    # Get the input vector
    F1, F2, d = u

    # Get the system response from the original dynamics
    xdot, ydot, thetadot, xddot, yddot, thddot = \
        pvtol_update(t, x, u[0:2], params)

    # Now add the wind term
    m = params.get('m', 4.)             # mass of aircraft
    xddot += d / m

    return np.array([xdot, ydot, thetadot, xddot, yddot, thddot])

windy_pvtol = ct.NonlinearIOSystem(
    windy_update, pvtol_output, name="windy_pvtol",
    states = [f'x{i}' for i in range(6)],
    inputs = ['F1', 'F2', 'd'],
    outputs = [f'x{i}' for i in range(6)]
)

#
# PVTOL dynamics with noise and disturbances
# 

def noisy_update(t, x, u, params):
    # Get the inputs
    F1, F2, Dx, Dy, Nx, Ny, Nth = u

    # Get the system response from the original dynamics
    xdot, ydot, thetadot, xddot, yddot, thddot = \
        pvtol_update(t, x, u[0:2], params)

    # Get the parameter values we need
    m = params.get('m', 4.)             # mass of aircraft
    J = params.get('J', 0.0475)         # inertia around pitch axis

    # Now add the disturbances
    xddot += Dx / m
    yddot += Dy / m

    return np.array([xdot, ydot, thetadot, xddot, yddot, thddot])

def noisy_output(t, x, u, params):
    F1, F2, dx, Dy, Nx, Ny, Nth = u
    return x + np.array([Nx, Ny, Nth, 0, 0, 0])

noisy_pvtol = ct.NonlinearIOSystem(
    noisy_update, noisy_output, name="noisy_pvtol",
    states = [f'x{i}' for i in range(6)],
    inputs = ['F1', 'F2'] + ['Dx', 'Dy'] + ['Nx', 'Ny', 'Nth'],
    outputs = pvtol.state_labels
)

# Add the linearitizations to the dynamics as additional methods
def noisy_pvtol_A(x, u, params={}):
    # Get the parameter values we need
    m = params.get('m', 4.)             # mass of aircraft
    J = params.get('J', 0.0475)         # inertia around pitch axis
    c = params.get('c', 0.05)           # damping factor (estimated)

    # Get the angle and compute sine and cosine
    theta = x[[2]]
    cth, sth = cos(theta), sin(theta)

    # Return the linearized dynamics matrix
    return np.array([
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0, (-u[0] * sth - u[1] * cth)/m, -c/m, 0, 0],
        [0, 0, ( u[0] * cth - u[1] * sth)/m, 0, -c/m, 0],
        [0, 0, 0, 0, 0, 0]
    ])
pvtol.A = noisy_pvtol_A

# Plot the trajectory in xy coordinates
def plot_results(t, x, u):
    # Set the size of the figure
    plt.figure(figsize=(10, 6))

    # Top plot: xy trajectory
    plt.subplot(2, 1, 1)
    plt.plot(x[0], x[1])
    plt.xlabel('x [m]')
    plt.ylabel('y [m]')
    plt.axis('equal')

    # Time traces of the state and input
    plt.subplot(2, 4, 5)
    plt.plot(t, x[1])
    plt.xlabel('Time t [sec]')
    plt.ylabel('y [m]')

    plt.subplot(2, 4, 6)
    plt.plot(t, x[2])
    plt.xlabel('Time t [sec]')
    plt.ylabel('theta [rad]')

    plt.subplot(2, 4, 7)
    plt.plot(t, u[0])
    plt.xlabel('Time t [sec]')
    plt.ylabel('$F_1$ [N]')

    plt.subplot(2, 4, 8)
    plt.plot(t, u[1])
    plt.xlabel('Time t [sec]')
    plt.ylabel('$F_2$ [N]')
    plt.tight_layout()

#
# Additional functions for testing
#

# Check flatness calculations
def _pvtol_check_flat(test_points=None, verbose=False):
    if test_points is None:
        # If no test points, use internal set
        mg = 9.8 * 4
        test_points = [
            ([0, 0, 0, 0, 0, 0], [0, mg]),
            ([1, 0, 0, 0, 0, 0], [0, mg]),
            ([0, 1, 0, 0, 0, 0], [0, mg]),
            ([1, 1, 0.1, 0, 0, 0], [0, mg]),
            ([0, 0, 0.1, 0, 0, 0], [0, mg]),
            ([0, 0, 0, 1, 0, 0], [0, mg]),
            ([0, 0, 0, 0, 1, 0], [0, mg]),
            ([0, 0, 0, 0, 0, 0.1], [0, mg]),
            ([0, 0, 0, 1, 1, 0.1], [0, mg]),
            ([0, 0, 0, 0, 0, 0], [1, mg]),
            ([0, 0, 0, 0, 0, 0], [0, mg + 1]),
            ([0, 0, 0, 0, 0, 0.1], [1, mg]),
            ([0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, mg + 1]),
        ]
    elif isinstance(test_points, tuple):
        # If we only got one test point, convert to a list
        test_points = [test_points]

    for x, u in test_points:
        x, u = np.array(x), np.array(u)
        flag = pvtol_flat_forward(x, u)
        xc, uc = pvtol_flat_reverse(flag)
        print(f'({x}, {u}): ', end='')
        if verbose:
            print(f'\n  flag: {flag}')
            print(f'  check: ({xc}, {uc}): ', end='')
        if np.allclose(x, xc) and np.allclose(u, uc):
            print("OK")
        else:
            print("ERR")
    



# genswitch_plot.m - run the Collins genetic switch model
# RMM, 24 Jan 07
#
# This file contains an example from FBS of a simple dynamical model
# of a genetic switch.  Plots time traces and a phase portrait using
# the python-control library.

import os

import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import odeint
from control import phase_plot, box_grid

# Simple model of a genetic switch

# This function implements the basic model of the genetic switch
# Parameters taken from Gardner, Cantor and Collins, Nature, 2000
def genswitch(y, t, mu=4, n=2):
    return mu/(1 + y[1]**n) - y[0], mu/(1 + y[0]**n) - y[1]

# Run a simulation from an initial condition
tim1 = np.linspace(0, 10, 100)
sol1 = odeint(genswitch, [1, 5], tim1)

# Extract the equilibrium points
mu = 4; n = 2           # switch parameters
eqpt = np.empty(3)
eqpt[0] = sol1[0, -1]
eqpt[1] = sol1[1, -1]
eqpt[2] = 0             # fzero(@(x) mu/(1+x^2) - x, 2)

# Run another simulation showing switching behavior
tim2 = np.linspace(11, 25, 100)
sol2 = odeint(genswitch, sol1[-1, :] + [2, -2], tim2)

# First plot out the curves that define the equilibria
u = np.linspace(0, 4.5, 46)
f = np.divide(mu, (1 + u**n))   # mu/(1 + u^n), element-wise

plt.figure(1); plt.clf()
plt.axis([0, 5, 0, 5])                 # box on;
plt.plot(u, f, '-', f, u, '--')         # 'LineWidth', AM_data_linewidth)
plt.legend(('z1, f(z1)', 'z2, f(z2)'))    # legend(lgh, 'boxoff')
plt.plot([0, 3], [0, 3], 'k-')          # 'LineWidth', AM_ref_linewidth)
plt.plot(eqpt[0], eqpt[1], 'k.', eqpt[1], eqpt[0], 'k.',
         eqpt[2], eqpt[2], 'k.')        # 'MarkerSize', AM_data_markersize*3)
plt.xlabel('z1, f(z2)')
plt.ylabel('z2, f(z1)')

# Time traces
plt.figure(3); plt.clf()  # subplot(221)
plt.plot(tim1, sol1[:, 0], 'b-', tim1, sol1[:, 1], 'g--')
# set(pl, 'LineWidth', AM_data_linewidth)
plt.plot([tim1[-1], tim1[-1] + 1],
         [sol1[-1, 0], sol2[0, 1]], 'ko:',
         [tim1[-1], tim1[-1] + 1], [sol1[-1, 1], sol2[0, 0]], 'ko:')
# set(pl, 'LineWidth', AM_data_linewidth, 'MarkerSize', AM_data_markersize)
plt.plot(tim2, sol2[:, 0], 'b-', tim2, sol2[:, 1], 'g--')
# set(pl, 'LineWidth', AM_data_linewidth)
plt.axis([0, 25, 0, 5])

plt.xlabel('Time {\itt} [scaled]')
plt.ylabel('Protein concentrations [scaled]')
plt.legend(('z1 (A)', 'z2 (B)'))  # 'Orientation', 'horizontal')
# legend(legh, 'boxoff')

# Phase portrait
plt.figure(2)
plt.clf()  # subplot(221)
plt.axis([0, 5, 0, 5])  # set(gca, 'DataAspectRatio', [1, 1, 1])
phase_plot(genswitch, X0=box_grid([0, 5, 6], [0, 5, 6]), T=10,
           timepts=[0.2, 0.6, 1.2])

# Add the stable equilibrium points
plt.plot(eqpt[0], eqpt[1], 'k.', eqpt[1], eqpt[0], 'k.',
         eqpt[2], eqpt[2], 'k.')       # 'MarkerSize', AM_data_markersize*3)

plt.xlabel('Protein A [scaled]')
plt.ylabel('Protein B [scaled]')       # 'Rotation', 90)

if 'PYCONTROL_TEST_EXAMPLES' not in os.environ:
    plt.show()



# exception.py - exception definitions for the control package
#
# Author: Richard M. Murray
# Date: 31 May 2010
#
# This file contains definitions of standard exceptions for the control package
#
# Copyright (c) 2010 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id$

class ControlSlycot(ImportError):
    """Exception for Slycot import.  Used when we can't import a function
    from the slycot package"""
    pass

class ControlDimension(ValueError):
    """Raised when dimensions of system objects are not correct"""
    pass

class ControlArgument(TypeError):
    """Raised when arguments to a function are not correct"""
    pass

class ControlMIMONotImplemented(NotImplementedError):
    """Function is not currently implemented for MIMO systems"""
    pass

class ControlNotImplemented(NotImplementedError):
    """Functionality is not yet implemented"""
    pass

# Utility function to see if slycot is installed
slycot_installed = None
def slycot_check():
    global slycot_installed
    if slycot_installed is None:
        try:
            import slycot
            slycot_installed = True
        except:
            slycot_installed = False
    return slycot_installed


# Utility function to see if pandas is installed
pandas_installed = None
def pandas_check():
    global pandas_installed
    if pandas_installed is None:
        try:
            import pandas
            pandas_installed = True
        except:
            pandas_installed = False
    return pandas_installed

# Utility function to see if cvxopt is installed
cvxopt_installed = None
def cvxopt_check():
    global cvxopt_installed
    if cvxopt_installed is None:
        try:
            import cvxopt
            cvxopt_installed = True
        except:
            cvxopt_installed = False
    return cvxopt_installed
